# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import qrl.generated.qrlwallet_pb2 as qrlwallet__pb2


class WalletAPIStub(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Wallet API
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.AddNewAddress = channel.unary_unary(
        '/qrl.WalletAPI/AddNewAddress',
        request_serializer=qrlwallet__pb2.AddNewAddressReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.AddNewAddressResp.FromString,
        )
    self.AddNewAddressWithSlaves = channel.unary_unary(
        '/qrl.WalletAPI/AddNewAddressWithSlaves',
        request_serializer=qrlwallet__pb2.AddNewAddressWithSlavesReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.AddNewAddressResp.FromString,
        )
    self.ListAddresses = channel.unary_unary(
        '/qrl.WalletAPI/ListAddresses',
        request_serializer=qrlwallet__pb2.ListAddressesReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.ListAddressesResp.FromString,
        )
    self.RemoveAddress = channel.unary_unary(
        '/qrl.WalletAPI/RemoveAddress',
        request_serializer=qrlwallet__pb2.RemoveAddressReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RemoveAddressResp.FromString,
        )
    self.IsValidAddress = channel.unary_unary(
        '/qrl.WalletAPI/IsValidAddress',
        request_serializer=qrlwallet__pb2.ValidAddressReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.ValidAddressResp.FromString,
        )
    self.EncryptWallet = channel.unary_unary(
        '/qrl.WalletAPI/EncryptWallet',
        request_serializer=qrlwallet__pb2.EncryptWalletReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.EncryptWalletResp.FromString,
        )
    self.LockWallet = channel.unary_unary(
        '/qrl.WalletAPI/LockWallet',
        request_serializer=qrlwallet__pb2.LockWalletReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.LockWalletResp.FromString,
        )
    self.UnlockWallet = channel.unary_unary(
        '/qrl.WalletAPI/UnlockWallet',
        request_serializer=qrlwallet__pb2.UnlockWalletReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.UnlockWalletResp.FromString,
        )
    self.GetRecoverySeeds = channel.unary_unary(
        '/qrl.WalletAPI/GetRecoverySeeds',
        request_serializer=qrlwallet__pb2.GetRecoverySeedsReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.GetRecoverySeedsResp.FromString,
        )
    self.GetWalletInfo = channel.unary_unary(
        '/qrl.WalletAPI/GetWalletInfo',
        request_serializer=qrlwallet__pb2.GetWalletInfoReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.GetWalletInfoResp.FromString,
        )
    self.RelayTransferTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelayTransferTxn',
        request_serializer=qrlwallet__pb2.RelayTransferTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTransferTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelayTransferTxnBySlave',
        request_serializer=qrlwallet__pb2.RelayTransferTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayMessageTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelayMessageTxn',
        request_serializer=qrlwallet__pb2.RelayMessageTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayMessageTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelayMessageTxnBySlave',
        request_serializer=qrlwallet__pb2.RelayMessageTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTokenTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelayTokenTxn',
        request_serializer=qrlwallet__pb2.RelayTokenTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTokenTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelayTokenTxnBySlave',
        request_serializer=qrlwallet__pb2.RelayTokenTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTransferTokenTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelayTransferTokenTxn',
        request_serializer=qrlwallet__pb2.RelayTransferTokenTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTransferTokenTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelayTransferTokenTxnBySlave',
        request_serializer=qrlwallet__pb2.RelayTransferTokenTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelaySlaveTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelaySlaveTxn',
        request_serializer=qrlwallet__pb2.RelaySlaveTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelaySlaveTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelaySlaveTxnBySlave',
        request_serializer=qrlwallet__pb2.RelaySlaveTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.ChangePassphrase = channel.unary_unary(
        '/qrl.WalletAPI/ChangePassphrase',
        request_serializer=qrlwallet__pb2.ChangePassphraseReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.ChangePassphraseResp.FromString,
        )
    self.GetTransactionsByAddress = channel.unary_unary(
        '/qrl.WalletAPI/GetTransactionsByAddress',
        request_serializer=qrlwallet__pb2.TransactionsByAddressReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.TransactionsByAddressResp.FromString,
        )
    self.GetTransaction = channel.unary_unary(
        '/qrl.WalletAPI/GetTransaction',
        request_serializer=qrlwallet__pb2.TransactionReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.TransactionResp.FromString,
        )
    self.GetBalance = channel.unary_unary(
        '/qrl.WalletAPI/GetBalance',
        request_serializer=qrlwallet__pb2.BalanceReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.BalanceResp.FromString,
        )
    self.GetOTS = channel.unary_unary(
        '/qrl.WalletAPI/GetOTS',
        request_serializer=qrlwallet__pb2.OTSReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.OTSResp.FromString,
        )
    self.GetHeight = channel.unary_unary(
        '/qrl.WalletAPI/GetHeight',
        request_serializer=qrlwallet__pb2.HeightReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.HeightResp.FromString,
        )
    self.GetBlock = channel.unary_unary(
        '/qrl.WalletAPI/GetBlock',
        request_serializer=qrlwallet__pb2.BlockReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.BlockResp.FromString,
        )
    self.GetBlockByNumber = channel.unary_unary(
        '/qrl.WalletAPI/GetBlockByNumber',
        request_serializer=qrlwallet__pb2.BlockByNumberReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.BlockResp.FromString,
        )
    self.GetAddressFromPK = channel.unary_unary(
        '/qrl.WalletAPI/GetAddressFromPK',
        request_serializer=qrlwallet__pb2.AddressFromPKReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.AddressFromPKResp.FromString,
        )
    self.GetNodeInfo = channel.unary_unary(
        '/qrl.WalletAPI/GetNodeInfo',
        request_serializer=qrlwallet__pb2.NodeInfoReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.NodeInfoResp.FromString,
        )


class WalletAPIServicer(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Wallet API
  """

  def AddNewAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddNewAddressWithSlaves(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListAddresses(self, request, context):
    """rpc AddAddressFromSeed(AddAddressFromSeedReq) returns (AddAddressFromSeedResp);

    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RemoveAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def IsValidAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def EncryptWallet(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def LockWallet(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UnlockWallet(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetRecoverySeeds(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetWalletInfo(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTransferTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTransferTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayMessageTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayMessageTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTokenTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTokenTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTransferTokenTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTransferTokenTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelaySlaveTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelaySlaveTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ChangePassphrase(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransactionsByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransaction(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBalance(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetOTS(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetHeight(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlock(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlockByNumber(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAddressFromPK(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNodeInfo(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_WalletAPIServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'AddNewAddress': grpc.unary_unary_rpc_method_handler(
          servicer.AddNewAddress,
          request_deserializer=qrlwallet__pb2.AddNewAddressReq.FromString,
          response_serializer=qrlwallet__pb2.AddNewAddressResp.SerializeToString,
      ),
      'AddNewAddressWithSlaves': grpc.unary_unary_rpc_method_handler(
          servicer.AddNewAddressWithSlaves,
          request_deserializer=qrlwallet__pb2.AddNewAddressWithSlavesReq.FromString,
          response_serializer=qrlwallet__pb2.AddNewAddressResp.SerializeToString,
      ),
      'ListAddresses': grpc.unary_unary_rpc_method_handler(
          servicer.ListAddresses,
          request_deserializer=qrlwallet__pb2.ListAddressesReq.FromString,
          response_serializer=qrlwallet__pb2.ListAddressesResp.SerializeToString,
      ),
      'RemoveAddress': grpc.unary_unary_rpc_method_handler(
          servicer.RemoveAddress,
          request_deserializer=qrlwallet__pb2.RemoveAddressReq.FromString,
          response_serializer=qrlwallet__pb2.RemoveAddressResp.SerializeToString,
      ),
      'IsValidAddress': grpc.unary_unary_rpc_method_handler(
          servicer.IsValidAddress,
          request_deserializer=qrlwallet__pb2.ValidAddressReq.FromString,
          response_serializer=qrlwallet__pb2.ValidAddressResp.SerializeToString,
      ),
      'EncryptWallet': grpc.unary_unary_rpc_method_handler(
          servicer.EncryptWallet,
          request_deserializer=qrlwallet__pb2.EncryptWalletReq.FromString,
          response_serializer=qrlwallet__pb2.EncryptWalletResp.SerializeToString,
      ),
      'LockWallet': grpc.unary_unary_rpc_method_handler(
          servicer.LockWallet,
          request_deserializer=qrlwallet__pb2.LockWalletReq.FromString,
          response_serializer=qrlwallet__pb2.LockWalletResp.SerializeToString,
      ),
      'UnlockWallet': grpc.unary_unary_rpc_method_handler(
          servicer.UnlockWallet,
          request_deserializer=qrlwallet__pb2.UnlockWalletReq.FromString,
          response_serializer=qrlwallet__pb2.UnlockWalletResp.SerializeToString,
      ),
      'GetRecoverySeeds': grpc.unary_unary_rpc_method_handler(
          servicer.GetRecoverySeeds,
          request_deserializer=qrlwallet__pb2.GetRecoverySeedsReq.FromString,
          response_serializer=qrlwallet__pb2.GetRecoverySeedsResp.SerializeToString,
      ),
      'GetWalletInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetWalletInfo,
          request_deserializer=qrlwallet__pb2.GetWalletInfoReq.FromString,
          response_serializer=qrlwallet__pb2.GetWalletInfoResp.SerializeToString,
      ),
      'RelayTransferTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTransferTxn,
          request_deserializer=qrlwallet__pb2.RelayTransferTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTransferTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTransferTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelayTransferTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayMessageTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelayMessageTxn,
          request_deserializer=qrlwallet__pb2.RelayMessageTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayMessageTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelayMessageTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelayMessageTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTokenTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTokenTxn,
          request_deserializer=qrlwallet__pb2.RelayTokenTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTokenTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTokenTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelayTokenTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTransferTokenTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTransferTokenTxn,
          request_deserializer=qrlwallet__pb2.RelayTransferTokenTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTransferTokenTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTransferTokenTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelayTransferTokenTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelaySlaveTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelaySlaveTxn,
          request_deserializer=qrlwallet__pb2.RelaySlaveTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelaySlaveTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelaySlaveTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelaySlaveTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'ChangePassphrase': grpc.unary_unary_rpc_method_handler(
          servicer.ChangePassphrase,
          request_deserializer=qrlwallet__pb2.ChangePassphraseReq.FromString,
          response_serializer=qrlwallet__pb2.ChangePassphraseResp.SerializeToString,
      ),
      'GetTransactionsByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransactionsByAddress,
          request_deserializer=qrlwallet__pb2.TransactionsByAddressReq.FromString,
          response_serializer=qrlwallet__pb2.TransactionsByAddressResp.SerializeToString,
      ),
      'GetTransaction': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransaction,
          request_deserializer=qrlwallet__pb2.TransactionReq.FromString,
          response_serializer=qrlwallet__pb2.TransactionResp.SerializeToString,
      ),
      'GetBalance': grpc.unary_unary_rpc_method_handler(
          servicer.GetBalance,
          request_deserializer=qrlwallet__pb2.BalanceReq.FromString,
          response_serializer=qrlwallet__pb2.BalanceResp.SerializeToString,
      ),
      'GetOTS': grpc.unary_unary_rpc_method_handler(
          servicer.GetOTS,
          request_deserializer=qrlwallet__pb2.OTSReq.FromString,
          response_serializer=qrlwallet__pb2.OTSResp.SerializeToString,
      ),
      'GetHeight': grpc.unary_unary_rpc_method_handler(
          servicer.GetHeight,
          request_deserializer=qrlwallet__pb2.HeightReq.FromString,
          response_serializer=qrlwallet__pb2.HeightResp.SerializeToString,
      ),
      'GetBlock': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlock,
          request_deserializer=qrlwallet__pb2.BlockReq.FromString,
          response_serializer=qrlwallet__pb2.BlockResp.SerializeToString,
      ),
      'GetBlockByNumber': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlockByNumber,
          request_deserializer=qrlwallet__pb2.BlockByNumberReq.FromString,
          response_serializer=qrlwallet__pb2.BlockResp.SerializeToString,
      ),
      'GetAddressFromPK': grpc.unary_unary_rpc_method_handler(
          servicer.GetAddressFromPK,
          request_deserializer=qrlwallet__pb2.AddressFromPKReq.FromString,
          response_serializer=qrlwallet__pb2.AddressFromPKResp.SerializeToString,
      ),
      'GetNodeInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetNodeInfo,
          request_deserializer=qrlwallet__pb2.NodeInfoReq.FromString,
          response_serializer=qrlwallet__pb2.NodeInfoResp.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'qrl.WalletAPI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))
